// CS.CPP - Implements a lock hierarchy for Win32
//
// Author: Jeff Claar

#define STRICT
#include <windows.h>
#include <tchar.h>
#include <yvals.h>
#pragma hdrstop
#include <stdio.h>
#include "cs.hpp"

DWORD Thread CLockLevel::sCurrentThreadLevel = 0;
DWORD CLockLevel::sAssignedLevels = 0;

bool CLockLevel::sLocksEnabled =
    CLockLevel::LocksEnabled();

bool CLockLevel::LocksEnabled()
{
    // Query the registry under:
    // HKLM\Software\LockLevel.
    // If the value Enabled (type DWORD)
    // is set to 1, then lock checking is
    // enabled.
    HKEY hReg;
    bool bEnabled = false;
    if (RegOpenKeyEx(HKEY_LOCAL_MACHINE,
        _T("Software\\LockLevel"),
        0,
        KEY_QUERY_VALUE,
        &hReg) == ERROR_SUCCESS)
    {
        DWORD dwVal, dwSize = sizeof(DWORD);
        DWORD dwType;
        if (RegQueryValueEx(hReg,
            _T("Enabled"),
            NULL,
            &dwType,
            (LPBYTE) &dwVal,
            &dwSize) == ERROR_SUCCESS)
        {
            if (dwType == REG_DWORD)
                bEnabled = (dwVal != 0);
        }
        RegCloseKey(hReg);
    }
    return bEnabled;
}

CLockLevel::CLockLevel(int level) :
 mLevel((level >= sizeof(sCurrentThreadLevel) * 8) ? -1 : level),
 mRefCount(0)
{
    // Verify that the passed-in lock level does not overflow
    // the size of sCurrentThreadLevel.
    if (sLocksEnabled && level >= sizeof(sCurrentThreadLevel) * 8)
    {
        TCHAR sMsg[80];
        wsprintf(sMsg,
            _T("Only %d lock levels are supported.")
            _T(" Disabling level checking.\n"),
            sizeof(sCurrentThreadLevel) * 8);
        OutputDebugString(sMsg);
    }

    // Verify that the lock level is not currently used.
    if (sLocksEnabled && mLevel != -1)
    {
        if (sAssignedLevels & this->LockBit())
        {
            OutputDebugString(_T("Level already used.\n"));
            DebugBreak();
        }
        sAssignedLevels |= this->LockBit();
    }
}

void CLockLevel::VerifyLockOK()
{
    if (sLocksEnabled && mLevel != -1 &&
        (sCurrentThreadLevel & this->LockBit()) == 0)
    {
        // Make sure that no lock higher than the current
        // level has already ben acquired.
        //if ( (sCurrentThreadLevel & ~(this->LockBit() - 1)) >
        //    this->LockBit())
        if (sCurrentThreadLevel > this->LockBit())
        {
            OutputDebugString(_T("Invalid lock!\n"));
            DebugBreak();
        }
    }
}

DWORD C_CSLockLevel::Enter(DWORD dwTimeout)
{
    // First, make sure that acquiring this lock
    // does not violate the lock hierarchy.
    this->VerifyLockOK();
    // Enter the critical section
    EnterCriticalSection(&mCS);
    // Update the state variables for the lock.
    CLockLevel::Enter(dwTimeout);
    return WAIT_OBJECT_0;
}

void C_CSLockLevel::Leave()
{
    // Update the state variables for the lock.
    // This is done before the lock is released
    // so that the modifications are thread-safe.
    CLockLevel::Leave();
    LeaveCriticalSection(&mCS);
}

DWORD C_MutexLockLevel::Enter(DWORD dwTimeout)
{
    this->VerifyLockOK();
    DWORD dwResult =WaitForSingleObject(mMutex, dwTimeout);
    // If the result is WAIT_TIMEOUT or WAIT_FAILED, the
    // lock was not acquired, so the state variables should
    // not be updated.
    if (dwResult != WAIT_TIMEOUT && dwResult != WAIT_FAILED)
    {
        CLockLevel::Enter(dwTimeout);
    }
    return dwResult;
}

void C_MutexLockLevel::Leave()
{
    CLockLevel::Leave();
    ReleaseMutex(mMutex);
}

DWORD C_MutexLockLevel::EnterMultiple(int count,
    C_MutexLockLevel** locks, 
    BOOL bWaitAll,
    DWORD dwTimeout)
{
    // Copy the handles into a temporary array
    // that can be used with WaitForMultipleObjects.
    HANDLE h[MAXIMUM_WAIT_OBJECTS];
    for (int i = 0; i < count; i++)
    {
        locks[i]->VerifyLockOK();
        h[i] = locks[i]->mMutex;
    }

    DWORD dwResult = WaitForMultipleObjects(count, h, bWaitAll, dwTimeout);

    if (dwResult != WAIT_FAILED && dwResult != WAIT_TIMEOUT)
    {
        // At least one lock was acquired. If bWaitAll
        // is TRUE, all locks were acquired, so all the
        // state variables must be updated. Otherwise,
        // only one was acquired, so only that lock
        // should be updated.
        if (bWaitAll)
        {
            for (i = 0; i < count; i++)
            {
                locks[i]->CLockLevel::Enter(dwTimeout);
            }
        }
        else
        {
            locks[dwResult - WAIT_OBJECT_0]->CLockLevel::Enter();
        }
    }

    return dwResult;
}

