// CS.HPP - Defines a lock hierarchy for Win32
//
// Author: Jeff Claar

#ifndef __CS_HPP__
#define __CS_HPP__

#include <yvals.h>  // for bool in VC4.2

#define Thread  __declspec(thread)

class CLockLevel
{
private:
    // Lock level for this lock.
    const int mLevel;
    // Number of times the current thread has
    // recursively acquired the lock.
    int mRefCount;

    // Currently assigned levels.
    static DWORD sAssignedLevels;
    // Currently held locks by a particular thread.
    static DWORD Thread sCurrentThreadLevel;

	// Variable for registry query to determine
	// if locks are enabled.
	static bool sLocksEnabled;
	static bool LocksEnabled();

	DWORD LockBit() const;
protected:
    // Function to verify that the lock level
    // of this lock is higher than any lock currently
    // held by the calling thread.
    void VerifyLockOK();
public:
    CLockLevel(int level = -1);
    virtual ~CLockLevel();

    // Pure virtual functions for entering and
    // leaving a lock. The base class uses these
    // functions to update sCurrentThreadLevel
    // and to verify that a given lock is allowed.
    virtual DWORD Enter(DWORD dwTimeout = INFINITE) = 0;
    virtual void Leave() = 0;
};

class C_CSLockLevel : public CLockLevel
{
private:
    CRITICAL_SECTION mCS;
public:
    C_CSLockLevel(int level = -1);
    ~C_CSLockLevel();

    DWORD Enter(DWORD dwTimeout = INFINITE);
    void Leave();
};

class C_MutexLockLevel : public CLockLevel
{
private:
    HANDLE mMutex;
public:
    C_MutexLockLevel(int level = -1,
        LPSECURITY_ATTRIBUTES lpsa = NULL,
        BOOL bInitialOwner = FALSE);
    ~C_MutexLockLevel();

    DWORD Enter(DWORD dwTimeout = INFINITE);
    void Leave();

    // Acquires multiple locks. This function
    // verifies that all passed in locks are
    // valid; that is, there is no lock
    // currently held at a higher lock level.
    static DWORD EnterMultiple(int count,
        C_MutexLockLevel** locks,
        BOOL bWaitAll,
        DWORD dwTimeout = INFINITE);
};


inline CLockLevel::~CLockLevel()
{
    if (sLocksEnabled && mLevel != -1)
    {
        // Free up the bit corresponding
        // to this level.
        sAssignedLevels &= ~this->LockBit();
    }
}

inline DWORD CLockLevel::Enter(DWORD dwTimeout)
{
    // If this is the first time the lock has been
    // entered by the current thread (mRefCount == 0),
    // then set the bit in sCurrentThreadLevel.
    if (sLocksEnabled && mLevel != -1 && ++mRefCount == 1)
    {
        sCurrentThreadLevel |= this->LockBit();
    }
    return WAIT_OBJECT_0;
}

inline void CLockLevel::Leave()
{
    // If --mRefCount == 0, then the lock is being freed.
    // Clear the bit in sCurrentThreadLevel in this case.
    // Note that at this point, the lock is still being
    // held, so mRefCount is implicitly thread-safe.
    if (sLocksEnabled && mLevel != -1 && --mRefCount == 0)
    {
        sCurrentThreadLevel &= ~this->LockBit();
    }
}

inline DWORD CLockLevel::LockBit() const
{
	return 1 << mLevel;
}

inline C_CSLockLevel::C_CSLockLevel(int level) :
 CLockLevel(level)
{
    InitializeCriticalSection(&mCS);
}

inline C_CSLockLevel::~C_CSLockLevel()
{
    DeleteCriticalSection(&mCS);
}

inline C_MutexLockLevel::C_MutexLockLevel(int level,
    LPSECURITY_ATTRIBUTES lpsa,
    BOOL bInitialOwner) :
 CLockLevel(level),
 mMutex(CreateMutex(lpsa, bInitialOwner, NULL))
{
}

inline C_MutexLockLevel::~C_MutexLockLevel()
{
    CloseHandle(mMutex);
}

#endif